@@ -59,7 +59,7 @@ static bool is_valid_colorspace(uint8_t colorspace)
 
 static bool read_plane(BitstreamRange* range,
                        std::shared_ptr<HeifPixelImage> image, heif_channel channel,
-                       int width, int height, int bit_depth)
+                       uint32_t width, uint32_t height, int bit_depth)
 {
   if (width <= 0 || height <= 0) {
     return false;
@@ -73,11 +73,11 @@ static bool read_plane(BitstreamRange* range,
   if (!image->add_plane(channel, width, height, bit_depth)) {
     return false;
   }
-  int stride;
+  uint32_t stride;
   uint8_t* plane = image->get_plane(channel, &stride);
   assert(stride >= width);
   auto stream = range->get_istream();
-  for (int y = 0; y < height; y++, plane += stride) {
+  for (uint32_t y = 0; y < height; y++, plane += stride) {
     assert(stream->read(plane, width));
   }
   return true;
@@ -85,7 +85,7 @@ static bool read_plane(BitstreamRange* range,
 
 static bool read_plane_interleaved(BitstreamRange* range,
                                    std::shared_ptr<HeifPixelImage> image, heif_channel channel,
-                                   int width, int height, int bit_depth, int comps)
+                                   uint32_t width, uint32_t height, int bit_depth, int comps)
 {
   if (width <= 0 || height <= 0) {
     return false;
@@ -99,11 +99,11 @@ static bool read_plane_interleaved(BitstreamRange* range,
   if (!image->add_plane(channel, width, height, bit_depth)) {
     return false;
   }
-  int stride;
+  uint32_t stride;
   uint8_t* plane = image->get_plane(channel, &stride);
   assert(stride >= width * comps);
   auto stream = range->get_istream();
-  for (int y = 0; y < height; y++, plane += stride) {
+  for (uint32_t y = 0; y < height; y++, plane += stride) {
     assert(stream->read(plane, width * comps));
   }
   return true;
@@ -114,8 +114,8 @@ extern "C" int LLVMFuzzerTestOneInput(const uint8_t* data, size_t size)
   auto reader = std::make_shared<StreamReader_memory>(data, size, false);
   BitstreamRange range(reader, size);
 
-  int width;
-  int height;
+  uint32_t width;
+  uint32_t height;
   int bit_depth;
   bool alpha;
   uint8_t in_chroma;
