@@ -18,6 +18,7 @@
  * along with libheif.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <cassert>
 #include "hdr_sdr.h"
 
 
@@ -30,7 +31,7 @@ Op_to_hdr_planes::state_after_conversion(const ColorState& input_state,
        input_state.chroma != heif_chroma_420 &&
        input_state.chroma != heif_chroma_422 &&
        input_state.chroma != heif_chroma_444) ||
-      input_state.bits_per_pixel != 8) {
+      input_state.bits_per_pixel != 8) { // TODO: support for <8 bpp
     return {};
   }
 
@@ -43,7 +44,7 @@ Op_to_hdr_planes::state_after_conversion(const ColorState& input_state,
   output_state = input_state;
   output_state.bits_per_pixel = target_state.bits_per_pixel;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -70,8 +71,8 @@ Op_to_hdr_planes::convert_colorspace(const std::shared_ptr<const HeifPixelImage>
                                heif_channel_B,
                                heif_channel_Alpha}) {
     if (input->has_channel(channel)) {
-      int width = input->get_width(channel);
-      int height = input->get_height(channel);
+      uint32_t width = input->get_width(channel);
+      uint32_t height = input->get_height(channel);
       if (!outimg->add_plane(channel, width, height, target_state.bits_per_pixel)) {
         return nullptr;
       }
@@ -80,20 +81,21 @@ Op_to_hdr_planes::convert_colorspace(const std::shared_ptr<const HeifPixelImage>
       int output_bits = target_state.bits_per_pixel;
 
       int shift1 = output_bits - input_bits;
-      int shift2 = 8 - shift1;
+      int shift2 = 2 * input_bits - output_bits;
 
       const uint8_t* p_in;
-      int stride_in;
+      uint32_t stride_in;
       p_in = input->get_plane(channel, &stride_in);
 
       uint16_t* p_out;
-      int stride_out;
+      uint32_t stride_out;
       p_out = (uint16_t*) outimg->get_plane(channel, &stride_out);
       stride_out /= 2;
 
-      for (int y = 0; y < height; y++)
-        for (int x = 0; x < width; x++) {
+      for (uint32_t y = 0; y < height; y++)
+        for (uint32_t x = 0; x < width; x++) {
           int in = p_in[y * stride_in + x];
+          // TODO: support for <8 bpp may need more than two copies of the input bit pattern
           p_out[y * stride_out + x] = (uint16_t) ((in << shift1) | (in >> shift2));
         }
     }
@@ -124,12 +126,12 @@ Op_to_sdr_planes::state_after_conversion(const ColorState& input_state,
 
   ColorState output_state;
 
-  // --- decrease bit depth
+  // --- output bit depth = 8
 
   output_state = input_state;
   output_state.bits_per_pixel = 8;
 
-  states.push_back({output_state, SpeedCosts_Unoptimized});
+  states.emplace_back(output_state, SpeedCosts_Unoptimized);
 
   return states;
 }
@@ -160,8 +162,8 @@ Op_to_sdr_planes::convert_colorspace(const std::shared_ptr<const HeifPixelImage>
       int input_bits = input->get_bits_per_pixel(channel);
 
       if (input_bits > 8) {
-        int width = input->get_width(channel);
-        int height = input->get_height(channel);
+        uint32_t width = input->get_width(channel);
+        uint32_t height = input->get_height(channel);
         if (!outimg->add_plane(channel, width, height, 8)) {
           return nullptr;
         }
@@ -169,21 +171,62 @@ Op_to_sdr_planes::convert_colorspace(const std::shared_ptr<const HeifPixelImage>
         int shift = input_bits - 8;
 
         const uint16_t* p_in;
-        int stride_in;
+        uint32_t stride_in;
         p_in = (uint16_t*) input->get_plane(channel, &stride_in);
         stride_in /= 2;
 
         uint8_t* p_out;
-        int stride_out;
+        uint32_t stride_out;
         p_out = outimg->get_plane(channel, &stride_out);
 
-        for (int y = 0; y < height; y++)
-          for (int x = 0; x < width; x++) {
+        for (uint32_t y = 0; y < height; y++)
+          for (uint32_t x = 0; x < width; x++) {
             int in = p_in[y * stride_in + x];
             p_out[y * stride_out + x] = (uint8_t) (in >> shift); // TODO: I think no rounding here, but am not sure.
           }
-      }
-      else {
+      } else if (input_bits < 8) {
+        uint32_t width = input->get_width(channel);
+        uint32_t height = input->get_height(channel);
+        if (!outimg->add_plane(channel, width, height, 8)) {
+          return nullptr;
+        }
+
+        // We also want to support converting inputs with < 4 bits per pixel covering the whole output range.
+        // E.g. a 1-bit input should map to the output 0x00 / 0xFF.
+        // We do so by constructing a fixed-point multiplication factor that effectively shifts and combines the input to
+        // a string of bits that completely fills the output.
+        //
+        // Example: input 3 bit.
+        // Factor (binary): 00100100|10010010
+        // Input copies:    AAABBBCC|CDDDEEE0
+        //                  \      /
+        //                   output
+
+        assert(input_bits > 0 && input_bits < 8);
+        auto bit = static_cast<uint16_t>(1 << (16 - input_bits));
+        uint16_t mulFactor = bit;
+
+        for (;;) {
+          bit >>= input_bits;
+          if (!bit) {
+            break;
+          }
+
+          mulFactor |= bit;
+        }
+
+        uint32_t stride_in;
+        const uint8_t* p_in = input->get_plane(channel, &stride_in);
+
+        uint32_t stride_out;
+        uint8_t* p_out = outimg->get_plane(channel, &stride_out);
+
+        for (uint32_t y = 0; y < height; y++)
+          for (uint32_t x = 0; x < width; x++) {
+            int in = p_in[y * stride_in + x];
+            p_out[y * stride_out + x] = (uint8_t) ((in * mulFactor) >> 8);
+          }
+      } else {
         outimg->copy_new_plane_from(input, channel, channel);
       }
     }
