@@ -33,9 +33,7 @@
 #include <sstream>
 
 #include "libheif/heif.h"
-
-
-static constexpr char kSuccess[] = "Success";
+#include <cassert>
 
 
 class ErrorBuffer
@@ -79,7 +77,7 @@ public:
         heif_suberror_code sc = heif_suberror_Unspecified,
         const std::string& msg = "");
 
-  static Error Ok;
+  static const Error Ok;
 
   static const char kSuccess[];
 
@@ -107,9 +105,21 @@ inline std::ostream& operator<<(std::ostream& ostr, const Error& err)
 template <typename T> class Result
 {
 public:
+  Result() = default;
+
+  Result(const T& v) : value(v), error(Error::Ok) {}
+
+  Result(const Error& e) : error(e) {}
+
   operator bool() const { return error.error_code == heif_error_Ok; }
 
-  T value;
+  T& operator*()
+  {
+    assert(error.error_code == heif_error_Ok);
+    return value;
+  }
+
+  T value{};
   Error error;
 };
 
